import shutil
from pathlib import Path

from bx_py_utils.path import assert_is_dir
from rich import print  # noqa
from rich.pretty import pprint

from manageprojects.git import Git


def iter_context(*, context: dict, prefix='') -> tuple:
    for key, value in context.items():
        if key.startswith('_'):
            continue

        if isinstance(value, dict):
            yield from iter_context(context=value, prefix=f'{key}.')
        else:
            yield (f'{prefix}{key}', value)


def generate_reverse_info(*, cookiecutter_context: dict) -> tuple:
    reverse_info = [
        (value, '{{ %s }}' % key) for key, value in iter_context(context=cookiecutter_context)
    ]
    reverse_info.sort(key=lambda x: len(x[0]), reverse=True)
    return tuple(reverse_info)


def replace_path(*, path: Path, reverse_info: tuple) -> Path:
    new_parts = []
    for part in path.parts:
        for src_str, dst_str in reverse_info:
            part = part.replace(src_str, dst_str)
        new_parts.append(part)
    return Path(*new_parts)


def build_dst_path(*, source_path: Path, item: Path, destination: Path, reverse_info: tuple):
    rel_path = item.relative_to(source_path)
    new_path = replace_path(path=rel_path, reverse_info=reverse_info)
    dst_path = destination / new_path
    return dst_path


def copy_replaced(src_path, dst_path, reverse_info):
    dst_parent = dst_path.parent
    dst_parent.mkdir(parents=True, exist_ok=True)

    try:
        content = src_path.read_text(encoding='UTF-8')
    except UnicodeDecodeError as err:
        # XXX: Detect binary files in a other way
        print(f'[yellow]Warning: {err} for file {src_path}')
        print('copy as binary file')
        shutil.copy2(src_path, dst_path)
    else:
        for src_str, dst_str in reverse_info:
            content = content.replace(src_str, dst_str)

        dst_path.write_text(content, encoding='UTF-8')


def create_cookiecutter_template(
    *,
    source_path: Path,
    destination: Path,
    cookiecutter_context: dict,
    overwrite: bool = False,
):
    source_path = source_path.resolve()
    assert_is_dir(source_path)
    if not overwrite:
        assert not destination.exists(), f'Destination {destination} already exists!'

    reverse_info = generate_reverse_info(cookiecutter_context=cookiecutter_context)

    print('Use this reverse context:')
    pprint(reverse_info)

    git = Git(cwd=source_path, detect_root=True)
    file_paths = git.ls_files(verbose=True)

    for item in file_paths:
        dst_path = build_dst_path(
            source_path=source_path,
            item=item,
            destination=destination,
            reverse_info=reverse_info,
        )
        if item.is_dir():
            dst_path.mkdir(parents=True, exist_ok=True)
        elif item.is_file():
            print(item.relative_to(source_path), '->', dst_path)
            copy_replaced(src_path=item, dst_path=dst_path, reverse_info=reverse_info)
        else:
            print(f'Ignore: {item}')



import inspect
import json
from pathlib import Path

from bx_py_utils.path import assert_is_dir
from bx_py_utils.test_utils.snapshot import assert_text_snapshot

from manageprojects.data_classes import GenerateTemplatePatchResult
from manageprojects.patching import generate_template_patch, make_git_diff
from manageprojects.test_utils.git_utils import init_git
from manageprojects.test_utils.logs import AssertLogs
from manageprojects.tests.base import BaseTestCase
from manageprojects.utilities.temp_path import TemporaryDirectory


class PatchingTestCase(BaseTestCase):
    def test_make_git_diff(self):
        with TemporaryDirectory(prefix='test_make_git_diff_') as main_temp_path:
            from_path = main_temp_path / 'from'
            to_path = main_temp_path / 'to'

            from_path.mkdir()
            to_path.mkdir()

            Path(from_path, 'file1.txt').write_text('Rev 1')
            Path(to_path, 'file1.txt').write_text('Rev 2')

            renamed_file_content = inspect.cleandoc(
                '''
                # This file is just renamed,
                # so the content is the same ;)
                '''
            )
            Path(from_path, 'old_name.txt').write_text(renamed_file_content)
            Path(to_path, 'new_name.txt').write_text(renamed_file_content)

            with AssertLogs(self, loggers=('manageprojects',)) as logs:
                patch = make_git_diff(
                    temp_path=main_temp_path,
                    from_path=from_path,
                    to_path=to_path,
                    verbose=True,
                )
            logs.assert_in('old_name.txt', 'new_name.txt')
            self.assertIn('diff --git a/file1.txt b/file1.txt', patch)
            self.assertIn('-Rev 1', patch)
            self.assertIn('+Rev 2', patch)
            self.assertIn('rename from old_name.txt', patch)
            self.assertIn('rename to new_name.txt', patch)
            assert_text_snapshot(got=patch, extension='.patch')

    def test_generate_template_patch(self):
        rev1_content = inspect.cleandoc(
            '''
            # This is a test line, not changed
            #
            # Revision 1
            #
            # The same cookiecutter value:
            print('Test: {{ cookiecutter.value }}')
            '''
        )
        rev2_content = inspect.cleandoc(
            '''
            # This is a test line, not changed
            #
            # Revision 2
            #
            # The same cookiecutter value:
            print('Test: {{ cookiecutter.value }}')
            '''
        )

        with TemporaryDirectory(prefix='test_generate_template_patch_') as main_temp_path:
            project_path = main_temp_path / 'main_temp_path'
            repo_path = main_temp_path / 'repo_path'

            test_file_path = (
                repo_path / '{{cookiecutter.dir_name}}' / '{{cookiecutter.file_name}}.py'
            )
            test_file_path.parent.mkdir(parents=True)
            test_file_path.write_text(rev1_content)
            cookiecutter_json_path = repo_path / 'cookiecutter.json'
            cookiecutter_json_path.write_text(
                json.dumps(
                    {
                        'dir_name': 'a_directory',
                        'file_name': 'a_file_name',
                        'value': 'FooBar',
                    }
                )
            )

            git, from_rev = init_git(repo_path)

            test_file_path.write_text(rev2_content)
            git.add('.', verbose=False)
            git.commit('The second commit', verbose=False)
            to_rev = git.get_current_hash(verbose=False)
            to_date = git.get_commit_date(verbose=False)

            patch_file_path = Path(
                project_path,
                '.manageprojects',
                'patches',
                f'{from_rev}_{to_rev}.patch',
            )
            self.assertFalse(patch_file_path.exists())

            with AssertLogs(self) as logs:
                result = generate_template_patch(
                    project_path=project_path,
                    template=str(repo_path),
                    directory=None,
                    from_rev=from_rev,
                    replay_context={},
                    cleanup=False,  # Keep temp files if this test fails, for better debugging
                    no_input=True,  # No user input in tests ;)
                )
            logs.assert_in("Call 'cookiecutter'", 'Write patch file')
            self.assertIsInstance(result, GenerateTemplatePatchResult)

            self.assertEqual(result.patch_file_path, patch_file_path)
            self.assertEqual(result.from_rev, from_rev)
            self.assertEqual(result.to_rev, to_rev)
            self.assertEqual(result.to_commit_date, to_date)

            self.assert_file_content(
                patch_file_path,
                inspect.cleandoc(
                    r'''
                    diff --git a/a_file_name.py b/a_file_name.py
                    index 4d0e75c..c71c9fe 100644
                    --- a/a_file_name.py
                    +++ b/a_file_name.py
                    @@ -1,6 +1,6 @@
                     # This is a test line, not changed
                     #
                    -# Revision 1
                    +# Revision 2
                     #
                     # The same cookiecutter value:
                     print('Test: FooBar')
                    \ No newline at end of file
                    '''
                ),
            )

            # The repro path contains the "from" revision:
            assert_is_dir(result.repo_path)
            self.assert_file_content(
                Path(
                    result.repo_path,
                    '{{cookiecutter.dir_name}}',
                    '{{cookiecutter.file_name}}.py',
                ),
                rev1_content,
            )



import logging
import subprocess
import sys
from pathlib import Path
from typing import Optional

from rich import print as rprint

from manageprojects.cookiecutter_api import execute_cookiecutter
from manageprojects.cookiecutter_generator import create_cookiecutter_template
from manageprojects.data_classes import (
    CookiecutterResult,
    GenerateTemplatePatchResult,
    ManageProjectsMeta,
    OverwriteResult,
)
from manageprojects.git import Git
from manageprojects.overwrite import overwrite_project
from manageprojects.patching import generate_template_patch
from manageprojects.utilities.pyproject_toml import PyProjectToml


logger = logging.getLogger(__name__)


def start_managed_project(
    *,
    template: str,  # CookieCutter Template path or GitHub url
    output_dir: Path,  # Target path where CookieCutter should store the result files
    directory: Optional[str] = None,  # Directory name of the CookieCutter Template
    input: bool = True,  # Prompt the user at command line for manual configuration?
    extra_context: Optional[dict] = None,
    replay: Optional[bool] = None,
    checkout: Optional[str] = None,
    password: Optional[str] = None,  # Optional password to use when extracting the repository
    config_file: Optional[Path] = None,  # Optional path to 'cookiecutter_config.yaml'
) -> CookiecutterResult:
    """
    Start a new "managed" project by run cookiecutter and create/update "pyproject.toml"
    """

    #############################################################################
    # Run cookiecutter

    cookiecutter_context, destination_path, repo_path = execute_cookiecutter(
        template=template,
        directory=directory,
        output_dir=output_dir,
        no_input=not input,
        extra_context=extra_context,
        replay=replay,
        checkout=checkout,
        password=password,
        config_file=config_file,
    )

    git = Git(cwd=repo_path)
    logger.debug('Cookiecutter git repro: %s', git.cwd)
    current_hash = git.get_current_hash()
    commit_date = git.get_commit_date()
    logger.info('Cookiecutter git repro: %s hash: %r %s', git.cwd, current_hash, commit_date)

    #############################################################################
    # Create or update "pyproject.toml" with manageprojects information:

    toml = PyProjectToml(project_path=destination_path)
    toml.init(
        revision=current_hash,
        dt=commit_date,
        template=template,
        directory=directory,
    )
    toml.create_or_update_cookiecutter_context(context=cookiecutter_context)
    toml.save()

    result = CookiecutterResult(
        destination_path=destination_path,
        git_path=git.cwd,
        git_hash=current_hash,
        commit_date=commit_date,
        cookiecutter_context=cookiecutter_context,
    )
    return result


def update_managed_project(
    project_path: Path,
    overwrite: bool = False,  # Don't apply git patches -> Just overwrite all template files!
    password: Optional[str] = None,
    config_file: Optional[Path] = None,  # CookieCutter config file
    cleanup: bool = True,  # Remove temp files if not exceptions happens
    input: bool = False,  # Prompt the user at command line for manual configuration?
) -> Optional[GenerateTemplatePatchResult]:
    """
    Update a existing project by apply git patch from cookiecutter template changes.
    """
    git = Git(cwd=project_path, detect_root=True)

    #############################################################################
    # Get information from project's pyproject.toml

    toml = PyProjectToml(project_path=project_path)
    meta: ManageProjectsMeta = toml.get_mp_meta()

    from_rev = meta.get_last_git_hash()
    assert from_rev, f'Fail to get last git hash from {toml.path}'

    cookiecutter_context = meta.cookiecutter_context
    assert cookiecutter_context, f'Missing cookiecutter context in {toml.path}'

    cookiecutter_template = meta.cookiecutter_template
    assert cookiecutter_template, f'Missing template in {toml.path}'

    if overwrite:
        # Don't apply git patches -> Just overwrite all template files:
        result = overwrite_project(
            git=git,
            project_path=project_path,
            template=cookiecutter_template,
            replay_context=cookiecutter_context,
            directory=meta.cookiecutter_directory,
            from_rev=from_rev,
            password=password,
            config_file=config_file,
            cleanup=cleanup,
            no_input=not input,
        )
        if not result:
            logger.info('Project is up-to-date, no changed to applied.')
            return None
        assert isinstance(result, OverwriteResult)
    else:
        # Generate the git diff/patch

        result = generate_template_patch(
            project_path=project_path,
            template=cookiecutter_template,
            directory=meta.cookiecutter_directory,
            from_rev=from_rev,
            replay_context=cookiecutter_context,
            password=password,
            config_file=config_file,
            cleanup=cleanup,
            no_input=not input,
        )
        if not result:
            logger.info('No git patch was created, nothing to apply.')
            return None
        assert isinstance(result, GenerateTemplatePatchResult)

        #############################################################################
        # Apply the patch

        patch_file_path = result.patch_file_path
        try:
            git.apply(patch_path=patch_file_path)
        except subprocess.CalledProcessError as err:
            print(err.stdout)
            if err.returncode == 1:
                print()
                print('Seems that the patch was not applied correctly!')
                print('Hint: run wiggle on the project:')
                print()
                print(f'./cli.py wiggle {project_path}')
                print()

    #############################################################################
    # Update "pyproject.toml" with applied patch information

    # Important: We *must* read the current "pyproject.toml" here again!
    # Otherwise, we may overwrite template changed with old content!
    toml = PyProjectToml(project_path=project_path)
    toml.add_applied_migrations(git_hash=result.to_rev, dt=result.to_commit_date)
    toml.save()

    return result


def clone_managed_project(
    project_path: Path,
    destination: Path,
    checkout: Optional[str] = None,  # Optional branch, tag or commit ID to checkout after clone
    password: Optional[str] = None,
    config_file: Optional[Path] = None,  # CookieCutter config file
    input: bool = False,  # Prompt the user at command line for manual configuration?
) -> CookiecutterResult:
    """
    Clone a existing project by replay the cookiecutter template in a new directory.
    """
    rprint(f'Clone {project_path} to {destination}')
    if destination.exists():
        print(f'ERROR: Destination {destination} already exists!')
        sys.exit(1)

    toml = PyProjectToml(project_path=project_path)
    meta: ManageProjectsMeta = toml.get_mp_meta()

    cookiecutter_context = meta.cookiecutter_context
    assert cookiecutter_context, f'Missing cookiecutter context in {toml.path}'

    extra_context = cookiecutter_context.get('cookiecutter')
    if not extra_context:
        print('WARNING: No "cookiecutter" in replay context!')

    cookiecutter_template = meta.cookiecutter_template
    assert cookiecutter_template, f'Missing template in {toml.path}'

    cookiecutter_directory = meta.cookiecutter_directory

    print('Use extra context:')
    rprint(extra_context)
    new_context, destination_path, repo_path = execute_cookiecutter(
        template=cookiecutter_template,
        directory=cookiecutter_directory,
        output_dir=destination,
        no_input=not input,
        extra_context=extra_context,
        checkout=checkout,
        password=password,
        config_file=config_file,
    )

    git = Git(cwd=repo_path)
    logger.debug('Cookiecutter git repro: %s', git.cwd)
    current_hash = git.get_current_hash()
    commit_date = git.get_commit_date()
    logger.info('Cookiecutter git repro: %s hash: %r %s', git.cwd, current_hash, commit_date)

    #############################################################################
    # Create "pyproject.toml" with manageprojects information:

    toml = PyProjectToml(project_path=destination_path)
    toml.init(
        revision=current_hash,
        dt=commit_date,
        template=cookiecutter_template,
        directory=cookiecutter_directory,
    )
    toml.create_or_update_cookiecutter_context(context=cookiecutter_context)
    toml.save()

    result = CookiecutterResult(
        destination_path=destination_path,
        git_path=git.cwd,
        git_hash=current_hash,
        commit_date=commit_date,
        cookiecutter_context=cookiecutter_context,
    )
    print(f'{project_path} successfully cloned to {destination_path}')
    return result


def reverse_managed_project(
    *,
    project_path: Path,
    destination: Path,
    overwrite: bool = False,
):
    """
    Create a cookiecutter template from a managed project.
    """
    rprint(f'Create cookiecutter template from {project_path} in {destination}')
    if not overwrite and destination.exists():
        print(f'ERROR: Destination {destination} already exists!')
        sys.exit(1)

    toml = PyProjectToml(project_path=project_path)
    meta: ManageProjectsMeta = toml.get_mp_meta()

    cookiecutter_context = meta.cookiecutter_context
    assert cookiecutter_context, f'Missing cookiecutter context in {toml.path}'

    create_cookiecutter_template(
        source_path=project_path,
        destination=destination,
        cookiecutter_context=cookiecutter_context,
        overwrite=overwrite,
    )



from unittest.mock import patch

import click
import rich_click
from bx_py_utils.path import assert_is_file
from bx_py_utils.test_utils.context_managers import MassContextManager
from click.testing import CliRunner, Result

from manageprojects.utilities.subprocess_utils import verbose_check_output


TERMINAL_WIDTH = 100


def subprocess_cli(*, cli_bin, args, exit_on_error=True):
    assert_is_file(cli_bin)
    return verbose_check_output(
        cli_bin,
        *args,
        cwd=cli_bin.parent,
        extra_env=dict(TERMINAL_WIDTH=str(TERMINAL_WIDTH)),
        exit_on_error=exit_on_error,
    )


class ForceRichTerminalWidth(MassContextManager):
    mocks = (
        patch.object(rich_click.rich_click, 'MAX_WIDTH', TERMINAL_WIDTH),
        patch.object(rich_click.rich_click, 'FORCE_TERMINAL', False),
    )


class ClickInvokeCliException(Exception):
    def __init__(self, result: Result):
        self.result = result


def invoke_click(cli, *args, expected_stderr='', expected_exit_code=0, strip=True, **kwargs):
    assert isinstance(cli, click.Command)

    args = tuple([str(arg) for arg in args])  # e.g.: Path() -> str

    with ForceRichTerminalWidth():
        runner = CliRunner(mix_stderr=False)
        result: Result = runner.invoke(cli=cli, args=args, **kwargs, color=False)

    if result.exception:
        raise ClickInvokeCliException(result) from result.exception

    stdout = result.stdout
    stderr = result.stderr
    assert stderr == expected_stderr, f'{stderr!r} is not {expected_stderr!r}'
    assert (
        result.exit_code == expected_exit_code
    ), f'Exit code {result.exit_code} is not {expected_exit_code}\n{stdout}\n{stderr}'

    if strip:
        stdout = '\n'.join(line.rstrip() for line in stdout.splitlines())
        stdout = stdout.strip()
    return stdout



import datetime
import filecmp
import inspect
import shutil
from pathlib import Path
from unittest import TestCase

from bx_py_utils.test_utils.datetime import parse_dt
from bx_py_utils.test_utils.snapshot import assert_text_snapshot
from packaging.version import Version

from manageprojects.cli.cli_app import PACKAGE_ROOT
from manageprojects.git import Git, GitTagInfo, GitTagInfos
from manageprojects.test_utils.git_utils import init_git
from manageprojects.test_utils.logs import AssertLogs
from manageprojects.utilities.temp_path import TemporaryDirectory


class GitTestCase(TestCase):
    def test_config(self):
        with TemporaryDirectory(prefix='test_init_git_') as temp_path:
            git = Git(cwd=temp_path, detect_root=False)
            git.init()

            self.assertEqual(git.get_config(key='user.name'), 'manageprojects')
            self.assertEqual(git.get_config(key='user.email'), 'manageprojects@test.tld')

            keys = git.list_config_keys()
            self.assertIsInstance(keys, set)
            self.assertGreaterEqual(len(keys), 1)

            section = 'manageprojects'
            test_key = f'{section}.test-config-entry'

            value = git.get_config(key=test_key)
            self.assertIsNone(value)

            git.config(key=test_key, value='test', scope='local')
            value = git.get_config(key=test_key)
            self.assertEqual(value, 'test')

            git.git_verbose_check_output('config', '--local', '--remove-section', section)

            value = git.get_config(key=test_key)
            self.assertIsNone(value)

    def test_own_git_repo(self):
        deep_path = Path(__file__).parent
        git = Git(cwd=deep_path)
        git_root_path = git.cwd
        self.assertNotEqual(git_root_path, deep_path)
        self.assertEqual(git_root_path, PACKAGE_ROOT)

        git_hash = git.get_current_hash(verbose=False)
        self.assertEqual(len(git_hash), 7, f'Wrong: {git_hash!r}')

        commit_date = git.get_commit_date(verbose=False)
        self.assertIsInstance(commit_date, datetime.datetime)
        self.assertGreater(commit_date, parse_dt('2023-01-01T00:00:00+0000'))
        self.assertLess(commit_date, parse_dt('2024-01-01T00:00:00+0000'))  # ;)

    def test_init_git(self):
        with TemporaryDirectory(prefix='test_init_git_') as temp_path:
            Path(temp_path, 'foo.txt').touch()
            Path(temp_path, 'bar.txt').touch()

            git, git_hash = init_git(temp_path)
            self.assertEqual(len(git_hash), 7)
            self.assertEqual(
                git.ls_files(verbose=False),
                [Path(temp_path, 'bar.txt'), Path(temp_path, 'foo.txt')],
            )

    def test_git_diff(self):
        with TemporaryDirectory(prefix='test_git_diff_') as temp_path:
            change_txt_path = Path(temp_path, 'change.txt')
            change_txt_path.write_text('This is the first revision!')
            Path(temp_path, 'unchange.txt').write_text('This file will be not changed')

            git, first_hash = init_git(temp_path)
            self.assertEqual(len(first_hash), 7)
            self.assertEqual(
                git.ls_files(verbose=False),
                [Path(temp_path, 'change.txt'), Path(temp_path, 'unchange.txt')],
            )

            change_txt_path.write_text('This is the second revision!')

            git.add('.', verbose=False)
            git.commit('The second commit', verbose=False)

            second_hash = git.get_current_hash(verbose=False)
            reflog = git.reflog(verbose=False)
            self.assertIn('The second commit', reflog)
            self.assertIn(first_hash, reflog)
            self.assertIn(second_hash, reflog)

            diff_txt = git.diff(first_hash, second_hash)
            self.assertIn('--- a/change.txt', diff_txt)
            self.assertIn('+++ b/change.txt', diff_txt)
            self.assertIn('@@ -1 +1 @@', diff_txt)
            assert_text_snapshot(got=diff_txt, extension='.patch')

    def test_git_apply_patch(self):
        with TemporaryDirectory(prefix='test_git_apply_patch_') as temp_path:
            repo_path = temp_path / 'git-repo'
            project_path = temp_path / 'project'

            repo_change_path = repo_path / 'directory1' / 'pyproject.toml'
            repo_change_path.parent.mkdir(parents=True)
            repo_change_path.write_text(
                inspect.cleandoc(
                    '''
                    [tool.darker]
                    src = ['.']
                    revision = "origin/main..."
                    line_length = 79  # 79 is initial, change to 100 later
                    verbose = true
                    diff = false
                    '''
                )
            )
            Path(repo_path, 'directory1', 'unchanged.txt').write_text('Will be not changed file')

            shutil.copytree(repo_path, project_path)  # "fake" cookiecutter output
            project_git, project_first_hash = init_git(project_path)  # init project

            # 1:1 copy?
            project_change_path = project_path / 'directory1' / 'pyproject.toml'
            self.assertTrue(filecmp.cmp(project_change_path, repo_change_path))

            # init "cookiecutter" source project:
            repo_git, repo_first_hash = init_git(repo_path)

            # Add a change to "fake" source:
            repo_change_path.write_text(
                inspect.cleandoc(
                    '''
                    [tool.darker]
                    src = ['.']
                    revision = "origin/main..."
                    line_length = 100  # 79 is initial, change to 100 later
                    verbose = true
                    diff = false
                    '''
                )
            )
            repo_git.add('.', verbose=False)
            repo_git.commit('The second commit', verbose=False)
            second_hash = repo_git.get_current_hash(verbose=False)

            # Generate a patch via git diff:
            diff_txt = repo_git.diff(repo_first_hash, second_hash)
            self.assertIn('directory1/pyproject.toml', diff_txt)
            patch_file_path = temp_path / 'git-diff-1.patch'
            patch_file_path.write_text(diff_txt)
            assert_text_snapshot(got=diff_txt, extension='.patch')

            # Change the project a little bit, before apply the git diff patch:

            # Just add a new file, unrelated to the diff patch:
            Path(project_path, 'directory1', 'new.txt').write_text('A new project file')

            # Commit the new file:
            project_git.add('.', verbose=False)
            project_git.commit('Add a new project file', verbose=False)

            # Change a diff related file:
            project_change_path.write_text(
                inspect.cleandoc(
                    '''
                    [tool.darker]
                    src = ['.']
                    revision = "origin/main..."
                    line_length = 79  # 79 is initial, change to 100 later
                    verbose = true
                    skip_string_normalization = true  # Added from project
                    diff = false
                    '''
                )
            )

            # Commit the changes, too:
            project_git.add('.', verbose=False)
            project_git.commit('Existing project file changed', verbose=False)

            # Now: Merge the project changes with the "fake" cookiecutter changes:
            project_git.apply(patch_file_path)

            # Merge successful?
            #  * line_length <<< change from "fake" cookiecutter changed
            #  * skip_string_normalization <<< added by project
            # The Rest is unchanged
            self.assertEqual(
                project_change_path.read_text(),
                inspect.cleandoc(
                    '''
                    [tool.darker]
                    src = ['.']
                    revision = "origin/main..."
                    line_length = 100  # 79 is initial, change to 100 later
                    verbose = true
                    skip_string_normalization = true  # Added from project
                    diff = false
                    '''
                ),
            )

    def test_status(self):
        with TemporaryDirectory(prefix='test_status_') as temp_path:
            change_txt_path = Path(temp_path, 'change.txt')
            change_txt_path.write_text('This is the first revision!')

            git, first_hash = init_git(temp_path)

            change_txt_path.write_text('Changed content')
            Path(temp_path, 'added.txt').write_text('Added file')

            status = git.status(verbose=False)
            self.assertEqual(status, [('M', 'change.txt'), ('??', 'added.txt')])

            git.add('.', verbose=False)
            git.commit('The second commit', verbose=False)

            status = git.status(verbose=False)
            self.assertEqual(status, [])

    def test_branch_names(self):
        with TemporaryDirectory(prefix='test_branch_names_') as temp_path:
            Path(temp_path, 'foo.txt').touch()
            git, first_hash = init_git(temp_path)

            with AssertLogs(self, loggers=('manageprojects',)) as logs:
                branch_names = git.get_branch_names()
                self.assertEqual(branch_names, ['main'])
            logs.assert_in("Git raw branches: ['* main']")

            git.git_verbose_check_call('checkout', '-b', 'foobar')

            with AssertLogs(self, loggers=('manageprojects',)) as logs:
                branch_names = git.get_branch_names()
                self.assertEqual(branch_names, ['foobar', 'main'])
            logs.assert_in("Git raw branches: ['* foobar', '  main']")

            with AssertLogs(self, loggers=('manageprojects',)):
                main_branch_name = git.get_main_branch_name()
                self.assertEqual(main_branch_name, 'main')

    def test_log(self):
        with TemporaryDirectory(prefix='test_get_version_from_tags') as temp_path:
            Path(temp_path, '1.txt').touch()
            git, first_hash = init_git(temp_path, comment='The initial commit ;)')

            git.tag('v0.0.1', message='one', verbose=False)

            Path(temp_path, '2.txt').touch()
            git.add(spec='.')
            git.commit(comment='Useless 1')

            Path(temp_path, '3.txt').touch()
            git.add(spec='.')
            git.commit(comment='Useless 2')

            git.tag('v0.2.0', message='two', verbose=False)
            Path(temp_path, '4.txt').touch()
            git.add(spec='.')
            git.commit(comment='Useless 3')

            output = git.log(format='%s')
            self.assertEqual(output, ['Useless 3', 'Useless 2', 'Useless 1', 'The initial commit ;)'])

            output = git.log(format='%s', no_merges=True, commit1='HEAD', commit2='v0.0.1')
            self.assertEqual(output, ['Useless 3', 'Useless 2', 'Useless 1'])

    def test_get_version_from_tags(self):
        with TemporaryDirectory(prefix='test_get_version_from_tags') as temp_path:
            Path(temp_path, 'foo.txt').touch()
            git, first_hash = init_git(temp_path)

            empty_tags = git.get_tag_infos()
            self.assertEqual(empty_tags, GitTagInfos(tags=[]))
            self.assertEqual(empty_tags.get_releases(), [])
            self.assertIs(empty_tags.get_last_release(), None)

            git.tag('v0.0.1', message='one', verbose=False)
            git.tag('v0.2.dev1', message='dev release', verbose=False)
            git.tag('v0.2.0a1', message='pre release', verbose=False)
            git.tag('v0.2.0rc2', message='release candidate', verbose=False)
            git.tag('v0.2.0', message='two', verbose=False)
            git.tag('foo', message='foo', verbose=False)
            git.tag('bar', message='bar', verbose=False)

            self.assertEqual(git.tag_list(), ['bar', 'foo', 'v0.0.1', 'v0.2.0', 'v0.2.0a1', 'v0.2.0rc2', 'v0.2.dev1'])

            with AssertLogs(self, loggers=('manageprojects',)) as logs:
                git_tag_infos = git.get_tag_infos()
                self.assertEqual(
                    git_tag_infos,
                    GitTagInfos(
                        tags=[
                            GitTagInfo(raw_tag='bar', version=None),
                            GitTagInfo(raw_tag='foo', version=None),
                            GitTagInfo(raw_tag='v0.0.1', version=Version('0.0.1')),
                            GitTagInfo(raw_tag='v0.2.dev1', version=Version('0.2.dev1')),
                            GitTagInfo(raw_tag='v0.2.0a1', version=Version('0.2.0a1')),
                            GitTagInfo(raw_tag='v0.2.0rc2', version=Version('0.2.0rc2')),
                            GitTagInfo(raw_tag='v0.2.0', version=Version('0.2.0')),
                        ]
                    ),
                )
            logs.assert_in('Ignore: Invalid version')

            self.assertTrue(git_tag_infos.exists(Version('0.0.1')))
            self.assertTrue(git_tag_infos.exists(Version('0.2.0rc2')))
            self.assertFalse(git_tag_infos.exists(Version('1.0')))

            self.assertEqual(
                git_tag_infos.get_releases(),
                [
                    GitTagInfo(raw_tag='v0.0.1', version=Version('0.0.1')),
                    GitTagInfo(raw_tag='v0.2.0', version=Version('0.2.0')),
                ],
            )
            self.assertEqual(
                git_tag_infos.get_last_release(),
                GitTagInfo(raw_tag='v0.2.0', version=Version('0.2.0')),
            )



import logging
from pathlib import Path
from unittest import TestCase


class AssertLogs:
    """
    Capture and assert log output from different loggers.
    """

    def __init__(
        self,
        test_case: TestCase,
        loggers: tuple[str] = ('manageprojects', 'cookiecutter'),
        level=logging.DEBUG,
    ):
        assertLogs = test_case.assertLogs

        self.logs = []
        for logger in loggers:
            self.logs.append(assertLogs(logger, level=level))

        self.context_managers = None

    def __enter__(self):
        self.context_managers = [log.__enter__() for log in self.logs]
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for log in self.logs:
            log.__exit__(exc_type, exc_val, exc_tb)

    def assert_in(self, *test_parts):
        outputs = []
        for cm in self.context_managers:
            outputs.extend(cm.output)

        output = '\n'.join(outputs)
        for part in test_parts:
            if isinstance(part, Path):
                part = str(part)
            assert part in output, f'Log part {part!r} not found in:\n{output}'
